/*
 
 File: ES1Renderer.m
 
 Abstract: The ES1Renderer class creates an OpenGL ES 1.1 context and draws 
 using OpenGL ES 1.1 functions.
 
 Version: 1.0
 
 Disclaimer: IMPORTANT:  This Apple software is supplied to you by Apple Inc.
 ("Apple") in consideration of your agreement to the following terms, and your
 use, installation, modification or redistribution of this Apple software
 constitutes acceptance of these terms.  If you do not agree with these terms,
 please do not use, install, modify or redistribute this Apple software.
 
 In consideration of your agreement to abide by the following terms, and subject
 to these terms, Apple grants you a personal, non-exclusive license, under
 Apple's copyrights in this original Apple software (the "Apple Software"), to
 use, reproduce, modify and redistribute the Apple Software, with or without
 modifications, in source and/or binary forms; provided that if you redistribute
 the Apple Software in its entirety and without modifications, you must retain
 this notice and the following text and disclaimers in all such redistributions
 of the Apple Software.
 Neither the name, trademarks, service marks or logos of Apple Inc. may be used
 to endorse or promote products derived from the Apple Software without specific
 prior written permission from Apple.  Except as expressly stated in this notice,
 no other rights or licenses, express or implied, are granted by Apple herein,
 including but not limited to any patent rights that may be infringed by your
 derivative works or by other works in which the Apple Software may be
 incorporated.
 
 The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
 WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
 WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
 COMBINATION WITH YOUR PRODUCTS.
 
 IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR
 DISTRIBUTION OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF
 CONTRACT, TORT (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF
 APPLE HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 Copyright (C) 2009 Apple Inc. All Rights Reserved.
 
*/

#import "ES1Renderer.h"

//Convert degrees to radians.
#define RADIANS(degrees)	((degrees) * M_PI / 180.0)

@implementation ES1Renderer

// Create an ES 1.1 context
- (id <ESRenderer>) init
{
	if (self = [super init])
	{
		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];
        
        if (!context || ![EAGLContext setCurrentContext:context])
		{
            [self release];
            return nil;
        }
		
		// Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
		glGenFramebuffersOES(1, &defaultFramebuffer);
		glGenRenderbuffersOES(1, &colorRenderbuffer);
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, colorRenderbuffer);
	}
	
	return self;
}

- (void) render {
    // Replace the implementation of this method to do your own custom drawing
	typedef GLfloat vertex_t[3];	//A vertex is an array of 3 floats.
	typedef vertex_t triangle_t[3];	//A triangle is an array of 3 vertices.
	
	//vertices on front face
#define A { 1.0f, -1.0f,  1.0f}
#define B { 1.0f,  1.0f,  1.0f}
#define C {-1.0f,  1.0f,  1.0f}
#define D {-1.0f, -1.0f,  1.0f}
	
	//vertices on rear face
#define E { 1.0f, -1.0f, -1.0f}
#define F { 1.0f,  1.0f, -1.0f}
#define G {-1.0f,  1.0f, -1.0f}
#define H {-1.0f, -1.0f, -1.0f}
	
#define FACE(a, b, c, d) {a, b, c}, {c, d, a}
	
	static const triangle_t cubeVertices[] = {
		FACE(A, B, C, D),	//front:  z ==  1
		FACE(E, F, G, H),	//rear:   z == -1
		
		FACE(B, F, G, C),	//top:    y ==  1
		FACE(A, D, H, E),	//bottom: y == -1
		
		FACE(A, E, F, B),	//right:  x ==  1
		FACE(C, G, H, D)	//left:   x == -1
	};
	
	typedef GLubyte color_t[4];	//A color is an array of 4 bytes: rgbα
	
#define RED    {255,   0,   0, 255}
#define GREEN  {  0, 255,   0, 255}
#define BLUE   {  0,   0, 255, 255}
	
#define YELLOW {255, 255,   0, 255}
#define CYAN   {  0, 255, 255, 255}
#define PURPLE {255,   0, 255, 255}
	
#define COLOR(c) c, c, c, c, c, c
	
	static const color_t cubeColors[] = {
		COLOR(RED),	//front
		COLOR(GREEN),	//rear
		
		COLOR(BLUE),	//top
		COLOR(YELLOW),	//bottom
		
		COLOR(CYAN),	//right
		COLOR(PURPLE)	//left
	};
    
	[EAGLContext setCurrentContext:context];
    
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);
	
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	
    //glOrthof(-1.0f, 1.0f, -1.5f, 1.5f, -1.0f, 1.0f);
	glEnable(GL_DEPTH_TEST);
	
	GLfloat near = .1;	//distance from viewer to near face of frustum
	//Horizontal field of view is 45 degrees.
	//length is half the horizontal length of the near face of the frustum.
	GLfloat length = near * tan(RADIANS(45.0f / 2.0f));
	GLfloat ratio = (GLfloat)backingHeight / backingWidth;
	//length = near/4;
	
	glFrustumf(
		-length,		//left
		length,			//right
		-length * ratio,	//bottom
		length * ratio,		//top
		near,			//near
		10.0			//far
	);
	
	glMatrixMode(GL_MODELVIEW);
	GLfloat orientation[4 * 4];
	glGetFloatv(GL_MODELVIEW_MATRIX, orientation);
	
	glPushMatrix();
	glLoadIdentity();
	glRotatef(-theta * 180.0 / M_PI, axis.x, axis.y, axis.z);
	
	glMultMatrixf(orientation);
	glGetFloatv(GL_MODELVIEW_MATRIX, orientation);

	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -5.0f);
	glMultMatrixf(orientation);
    
    glClearColor(0.5f, 0.5f, 0.5f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glVertexPointer(3, GL_FLOAT, 0, cubeVertices);
    glEnableClientState(GL_VERTEX_ARRAY);
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, cubeColors);
    glEnableClientState(GL_COLOR_ARRAY);
    glDrawArrays(GL_TRIANGLES, 3, 6 * 6 - 3);

	glPopMatrix();   
	glLoadMatrixf(orientation);
    
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

- (void) rotate: (float) t axis: (CGPoint3) p {
	theta = t;
	axis = p;
	[self render];
}

- (BOOL) resizeFromLayer:(CAEAGLLayer *)layer
{	
	// Allocate color buffer backing based on the current layer size
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:layer];
	glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES,
		&backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES,
		&backingHeight);
	
	glGenRenderbuffersOES(1, &depthRenderbuffer);
	glBindRenderbufferOES(GL_RENDERBUFFER_OES, depthRenderbuffer);
	
	glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES,
		backingWidth, backingHeight);
	
	glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES,
		GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depthRenderbuffer);
	
    if (glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES)
		!= GL_FRAMEBUFFER_COMPLETE_OES)
	{
		NSLog(@"Failed to make complete framebuffer object %x",
			glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }
    
    return YES;
}

- (void) dealloc
{
	// Tear down GL
	if (defaultFramebuffer)
	{
		glDeleteFramebuffersOES(1, &defaultFramebuffer);
		defaultFramebuffer = 0;
	}
	
	if (colorRenderbuffer)
	{
		glDeleteRenderbuffersOES(1, &colorRenderbuffer);
		colorRenderbuffer = 0;
	}
	
	// Tear down context
	if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];
	
	[context release];
	context = nil;
	
	[super dealloc];
}

@end
